class AmqpeekException(Exception):
    """Base exception for library"""
    pass


class ConfigExistsError(AmqpeekException):
    """
    Error thown when attempting to create a
    configuration file, but it already exists
    """



"""
Connecting to, and monitoring of RMQ
"""
import logging
import time

from pika import PlainCredentials, BlockingConnection, ConnectionParameters
from pika.exceptions import AMQPConnectionError


class Connector(object):
    """
    Abstracted connection to RMQ.

    Holds credentials for continued connection drop and reconnect
    """
    def __init__(self, host, port, vhost, user, passwd):
        """
        :param host: string
        :param port: string
        :param vhost: string
        :param user: string
        :param passwd: string
        """
        self.host = host
        self.port = port
        self.vhost = vhost
        self.user = user
        self.passwd = passwd

    def connect(self):
        """
        Create blocking connection in RMQ

        :return: pika.BlockingConnection
        """
        return BlockingConnection(
            parameters=ConnectionParameters(
                host=self.host,
                port=self.port,
                virtual_host=self.vhost,
                credentials=PlainCredentials(
                    username=self.user,
                    password=self.passwd,
                )
            )
        )


class Monitor(object):
    """
    Handles connection to RMQ, test of queues and sending notifications
    """
    def __init__(
        self, connector, queue_details, interval=None, max_connections=None
    ):
        """
        :param connector: amqpeek.monitor.Connector
        :param queue_details: dict
        :param interval: float
        :param max_connections: int
        """
        self.connector = connector
        self.queue_details = queue_details
        self.interval = interval
        self.max_connections = max_connections
        self.connection_count = 0
        self.notifiers = []

    def add_notifier(self, notifier):
        """
        :param notifier: amqpeek.notifier.Notifier
        """
        self.notifiers.append(notifier)

    def run(self):
        """
        Main execution loop
        """
        while True:
            try:
                connection = self.connector.connect()
            except AMQPConnectionError:
                subject = 'Connection Error'
                message = 'Error connecting to host: "{host}"'.format(
                    host=self.connector.host
                )

                logging.info('%s - %s', subject, message)
                self.notify(subject, message)
            else:
                self.check_queues(connection, self.queue_details)
                connection.close()

            if self.interval is not None:
                time.sleep(self.interval * 60)
                self.connection_count += 1

                if self.connection_count == self.max_connections:
                    return
            else:
                return

    def check_queues(self, connection, queue_details):
        """
        Check number of messages in queues are within limits

        :param connection: pika.BlockingConnection
        :param queue_details:
        """
        channel = self.get_channel(connection)

        for queue_name, queue_config in queue_details.items():
            queue = self.connect_to_queue(channel, queue_name, queue_config)
            message_count = self.get_queue_message_count(queue)

            if message_count > queue_config['limit']:
                subject = 'Queue Length Error'
                message = (
                    'Queue "{queue}" is over specified limit!! '
                    '({message_count} > {limit})'
                ).format(
                    queue=queue_name,
                    message_count=message_count,
                    limit=queue_config['limit']
                )

                logging.info('%s - %s', subject, message)
                self.notify(subject, message)

    def notify(self, subject, message):
        """
        :param subject: string
        :param message: string
        """
        for notifier in self.notifiers:
            notifier.notify(subject, message)

    def connect_to_queue(self, channel, queue_name, queue_config):
        """
        :param: channel: pika.channel.Channel
        :param: queue_name: string
        :param: queue_config: dict
        :return: pika.frame.Method
        """
        return channel.queue_declare(
            queue=queue_name,
            **queue_config['settings']
        )

    def get_channel(self, connection):
        """
        :param: connection: pika.BlockingConnection
        :return: pika.channel.Channel
        """
        return connection.channel()

    def get_queue_message_count(self, queue):
        """
        :param: queue: pika.frame.Method
        :return: int
        """
        return queue.method.message_count  # pragma: no cover



"""
Classes for connecting to different channels to send
notifications
"""
from smtplib import SMTP

from slacker import Slacker


def create_notifiers(notifier_data):
    """
    :param notifier_data: dict
    :return: tuple
    """
    if not notifier_data:
        return tuple()

    return tuple(
        NOTIFIER_MAP[notifier_type](**kwargs)
        for notifier_type, kwargs in notifier_data.items()
    )


class Notifier(object):
    """
    Base Notifier class
    """
    def notify(self, subject, message):
        """
        Send notifications
        """
        pass  # pragma: no cover


class SmtpNotifier(Notifier):
    """
    Sends Notifications via SMTP
    """
    MAIL_TEMPLATE = """\
    From: {from_addr}
    To: {to_addr}
    Subject: {subject}

    {message}
    """

    def __init__(
        self, host, to_addr, from_addr, subject, user=None, passwd=None
    ):
        """
        :param host: string
        :param user: string
        :param passwd: string
        :param to_addr: string
        :param from_addr: string
        :param subject: string
        :param user: string
        :param: passwd
        """
        self.host = host
        self.to_addr = to_addr
        self.from_addr = from_addr
        self.subject = subject
        self.user = user
        self.passwd = passwd
        self.server = SMTP(self.host)

        if self.user is not None:
            self.server.login(self.user, self.passwd)

    def notify(self, subject, message):
        """
        :param subject: string
        :param message: string
        """
        subject = "{base_subject} - {subject}".format(
            base_subject=self.subject,
            subject=subject
        )

        mail_message = self.MAIL_TEMPLATE.format(
            from_addr=self.from_addr,
            to_addr=", ".join(self.to_addr),
            subject=subject,
            message=message
        )

        self.server.sendmail(
            self.from_addr,
            self.to_addr,
            mail_message
        )


class SlackNotifier(Notifier):
    """
    Send notifications via Slack
    """
    def __init__(self, api_key, username, channel):
        """
        :param api_key: string
        :param username: string
        :param channel: string
        """
        self.username = username
        self.channel = channel
        self.slack = Slacker(api_key)

    def notify(self, subject, message):
        """
        :param subject: string
        :param message: string
        """
        message = "{subject}: {message}".format(
            subject=subject,
            message=message
        )

        self.slack.chat.post_message(
            channel=self.channel,
            text=message,
            username=self.username
        )


NOTIFIER_MAP = {
    'smtp': SmtpNotifier,
    'slack': SlackNotifier,
}



"""
CLI script and main entry point for amqpeek
"""
import logging
import os
import sys
from shutil import copyfile

import click
import yaml

from .exceptions import ConfigExistsError
from .notifier import create_notifiers
from .monitor import Connector, Monitor

DEFAULT_CONFIG = 'amqpeek.yaml'


def gen_config_file():
    if os.path.exists(DEFAULT_CONFIG):
        raise ConfigExistsError('File already exists')

    this_file = os.path.dirname(os.path.realpath(__file__))
    config_file = '{0}/../config/amqpeek.yaml'.format(this_file)
    copyfile(config_file, DEFAULT_CONFIG)


def read_config(config):
    """
    :param config: string
    :return: dict
    """
    with open(config, 'r') as config_file:
        return yaml.load(config_file)


def configure_logging(verbosity):
    """
    :param verbosity: int
    """
    if verbosity > 0:
        logging.basicConfig(format="%(message)s", level=logging.INFO)
    else:
        logging.basicConfig(format="%(message)s", level=logging.ERROR)


@click.command()
@click.option(
    '--config',
    '-c',
    default=DEFAULT_CONFIG,
    help=(
        'Location of configuration file to use '
        '(defaults to "amqpeek.yaml" in current directory)'
    )
)
@click.option(
    '--interval',
    '-i',
    type=float,
    default=None,
    help='Time interval between tests (minutes)'
)
@click.option(
    '--verbosity',
    '-v',
    count=True,
    default=0,
    help='Increase verbosity'
)
@click.option(
    '--max_tests',
    '-m',
    type=int,
    default=None,
    help='Maximum number of tests to run before stopping'
)
@click.option(
    '--gen_config',
    '-g',
    is_flag=True,
    help=(
        'Create a basic configuration file and place it in your '
        'current directory'
    )
)
def main(config, interval, verbosity, max_tests, gen_config):
    """
    AMQPeek - Simple, flexible RMQ monitor
    """
    configure_logging(verbosity)

    if gen_config:
        try:
            gen_config_file()
        except ConfigExistsError:
            click.echo(
                click.style(
                    'An AMQPeek config already exists in the current '
                    'directory',
                    fg='red'
                )
            )
        else:
            click.echo(
                click.style(
                    'AMQPeek config created',
                    fg='green'
                )
            )

            click.echo(
                click.style(
                    'Edit the file with your details and settings '
                    'before running AMQPeek',
                    fg='green'
                )
            )

        sys.exit(0)

    try:
        app_config = read_config(config)
    except IOError:
        click.echo(
            click.style(
                'No configuration file found. '
                'Specify a configuration file with --config. '
                'To generate a base config file use --gen-config.',
                fg='red'
            )
        )

        sys.exit(0)

    connector = Connector(**app_config['rabbit_connection'])

    monitor = Monitor(
        connector=connector,
        queue_details=app_config['queues'],
        interval=interval,
        max_connections=max_tests
    )

    notifiers = create_notifiers(app_config['notifiers'])

    for notifier in notifiers:
        monitor.add_notifier(notifier)

    monitor.run()



#!/usr/bin/env python
from codecs import open
import os
from setuptools import setup, find_packages


here = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(here, 'README.rst'), 'r', 'utf-8') as handle:
    readme = handle.read()

setup(
    name='amqpeek',
    description='RMQ CLI Monitor',
    packages=find_packages(exclude=['test', 'test.*']),
    version='0.0.2',
    long_description=readme,
    license="MIT",
    author='Steve Hutchins',
    author_email='hutchinsteve@gmail.com',
    url='https://github.com/steveYeah/AmqPeek',
    download_url='https://github.com/steveYeah/amqpeek/archive/v0.0.2.tar.gz',
    keywords=(
        'amqp',
        'monitor',
        'rabbitmq',
        'rabbit',
        'notifier',
        'notifications'
    ),
    install_requires=[
        'click>=6.6',
        'PyYAML>=3.12',
        'pika>=0.10.0',
        'slacker>=0.9.24',
    ],
    extras_require={
        'dev': [
            "pytest",
            "pep8",
            "mock==2.0.0",
        ],
    },
    entry_points={
        'console_scripts': [
            'amqpeek=amqpeek.cli:main',
        ],
    },
    package_dir={'amqpeek': 'amqpeek'},
    package_data={
        'amqpeek': ['../config/*.yaml']
    },
    classifiers=[
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'Topic :: System :: Systems Administration',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)
